
package com.atlassian.jira.plugin.ext.bazaar;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.collections.map.LRUMap;
import org.apache.log4j.Logger;
import org.vcs.bazaar.client.BazaarClientFactory;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.IBazaarClient;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.commandline.commands.options.Option;
import org.vcs.bazaar.client.commandline.syntax.ILogOptions;
import org.vcs.bazaar.client.core.BazaarClientException;
import org.vcs.bazaar.client.core.BranchLocation;

import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;

import com.atlassian.jira.InfrastructureException;
import com.atlassian.jira.plugin.ext.bazaar.linkrenderer.BazaarLinkRenderer;
import com.atlassian.jira.plugin.ext.bazaar.linkrenderer.LinkFormatRenderer;
import com.atlassian.jira.plugin.ext.bazaar.linkrenderer.NullLinkRenderer;
import com.opensymphony.module.propertyset.PropertySet;

/**
 * 
 * @author Guillermo Gonzalez <guillo.gonzo@gmail.com>
 *
 */
public class BazaarManagerImpl implements BazaarManager {
	private static Logger log = Logger.getLogger(BazaarManagerImpl.class);

	private BazaarLinkRenderer linkRenderer;
	private Map<BazaarRevision, IBazaarLogMessage> logEntryCache;
	private IBazaarClient repository;

	private boolean active;
	private String inactiveMessage;
	private long id;

	private PropertySet properties;

	private ViewLinkFormat viewLinkFormat = null;
	private boolean isViewLinkSet = false;

	public BazaarManagerImpl(long id, PropertySet props) {
		this.id = id;
		this.properties = props;
		setup();
	}

	public synchronized void update(BzrProperties props) {
		deactivate(null);

		BzrProperties.Util.fillPropertySet(props, properties);
		isViewLinkSet = false; /* If we don't reset this flag, we get SVN-190 */

		setup();
	}

	protected void setup() {
		// Now setup web link renderer
		linkRenderer = null;

		if (getViewLinkFormat() != null)
			linkRenderer = new LinkFormatRenderer(this);
		else
			linkRenderer = new NullLinkRenderer();

		// Now setup revision indexing if they want it
		if (isRevisionIndexing()) {
			// Setup the log message cache
			int cacheSize = 10000;

			if (getRevisioningCacheSize() > 0) {
				cacheSize = getRevisioningCacheSize();
			}

			logEntryCache = new LRUMap(cacheSize);
		}
		activate();
	}

	public synchronized Collection<IBazaarLogMessage> getLogEntries(
			BazaarRevision revision) {
		final Collection<IBazaarLogMessage> logEntries = new ArrayList<IBazaarLogMessage>();

		// if connection isn't up, don't even try
		if (!isActive()) {
			return logEntries;
		}

		// BazaarRevision latestRevision;
		// try {
		// latestRevision = repository.lo
		// }
		// catch (BazaarClientException e) {
		// // connection was active, but apparently now it's not
		// log.error("Error getting the latest revision from the repository.",
		// e);
		// deactivate(e.getMessage());
		// return logEntries;
		// }
		//
		// if (log.isDebugEnabled()) {
		// log.debug("Latest revision in repository=" + getRoot() + "  is : " +
		// latestRevision);
		// }
		//
		// if (latestRevision != null && latestRevision.compareTo(revision) <=
		// 0) {
		// if (log.isDebugEnabled()) {
		// log.debug("Have all the commits for repository=" + getRoot() +
		// " - doing nothing.");
		// }
		// return logEntries;
		// }
		//
		// long retrieveStart = revision + 1;
		// if (retrieveStart < 0) {
		// retrieveStart = 0;
		// }

		if (log.isDebugEnabled()) {
			log.debug("Retrieving revisions to index (between "
					+ revision.getValue() + " and last:) for branch="
					+ getRoot());
		}

		try {
			List<Option> options = new ArrayList<Option>(0);
			options.add(ILogOptions.VERBOSE);
			if (!BazaarRevision.INVALID.equals(revision)) {
				// IBazaarRevisionSpec revisions =
				// BazaarRevisionRange.getRange(revision,
				// BazaarRevision.getRevision(BazaarRevision.Prefix.LAST, "1"));
				options.add(ILogOptions.REVISION.setArgument(revision
						.getValue()
						+ ".." + "last:1"));
			}
			List<IBazaarLogMessage> logs = repository.log(new BranchLocation(URI
					.create(getRoot())), (Option[]) options
					.toArray(new Option[0]));
			for (IBazaarLogMessage logEntry : logs) {
				if (log.isDebugEnabled()) {
					log.debug("Retrieved #" + logEntry.getRevision() + " : "
							+ logEntry.getMessage());
				}
				logEntries.add(logEntry);
			}
		} catch (BazaarClientException e) {
			log.error("Error retrieving changes from the repository.", e);
			deactivate(e.getMessage());
		}
		if (log.isDebugEnabled()) {
			log.debug("Retrieved " + logEntries.size()
					+ " relevant revisions to index (between " + revision
					+ " and last:) from branch=" + getRoot());
		}

		// temp log comment
		if (log.isDebugEnabled()) {
			log.debug("log entries size = " + logEntries.size() + " for "
					+ getRoot());
		}
		return logEntries;
	}

	public synchronized IBazaarLogMessage getLogEntry(BazaarRevision revision) {
		if (!isActive()) {
			throw new IllegalStateException(
					"The connection to the repository is not active");
		}
		final IBazaarLogMessage[] logEntry = new IBazaarLogMessage[] { (IBazaarLogMessage) logEntryCache
				.get(revision) };

		if (logEntry[0] == null) {
			try {
				if (log.isDebugEnabled()) {
					log
							.debug("No cache - retrieving log message for revision: "
									+ revision);
				}

				List<IBazaarLogMessage> logs = repository.log(new BranchLocation(
						getRoot()), ILogOptions.REVISION.setArgument(revision
						.toString()), ILogOptions.VERBOSE);
				for (IBazaarLogMessage log : logs) {
					logEntry[0] = log;
					ensureCached(log);
				}
			} catch (BazaarClientException e) {
				log.error("Error retrieving logs: " + e, e);
				deactivate(e.getMessage());
				throw new InfrastructureException(e);
			}
		} else if (log.isDebugEnabled()) {
			log.debug("Found cached log message for revision: " + revision);
		}
		return logEntry[0];
	}

	public long getId() {
		return id;
	}

	/**
	 * Make sure a single log message is cached.
	 */
	private void ensureCached(IBazaarLogMessage logEntry) {
		synchronized (logEntryCache) {
			logEntryCache.put(logEntry.getRevision(), logEntry);
		}
	}

	public PropertySet getProperties() {
		return properties;
	}

	public String getDisplayName() {
		return !properties
				.exists(MultipleBazaarRepositoryManager.BZR_REPOSITORY_NAME) ? getRoot()
				: properties
						.getString(MultipleBazaarRepositoryManager.BZR_REPOSITORY_NAME);
	}

	public String getRoot() {
		return properties
				.getString(MultipleBazaarRepositoryManager.BZR_BRANCH_ROOT_KEY);
	}

	public String getUsername() {
		return properties
				.getString(MultipleBazaarRepositoryManager.BZR_USERNAME_KEY);
	}

	public String getPassword() {
		try {
			return decryptPassword(properties
					.getString(MultipleBazaarRepositoryManager.BZR_PASSWORD_KEY));
		} catch (IOException e) {
			log.error("Couldn't decrypt the password. Reseting it to null.", e);
			return null;
		}
	}

	public boolean isRevisionIndexing() {
		return properties
				.getBoolean(MultipleBazaarRepositoryManager.BZR_REVISION_INDEXING_KEY);
	}

	public int getRevisioningCacheSize() {
		return properties
				.getInt(MultipleBazaarRepositoryManager.BZR_REVISION_CACHE_SIZE_KEY);
	}

	public String getPrivateKeyFile() {
		return properties
				.getString(MultipleBazaarRepositoryManager.BZR_PRIVATE_KEY_FILE);
	}

	public String getXmlRpcPort() {
		return properties
				.getString(MultipleBazaarRepositoryManager.BZR_XMLRPC_PORT);
	}

	public String getBzrExecutable() {
		return properties
				.getString(MultipleBazaarRepositoryManager.BZR_EXECUTABLE);
	}

	public boolean isActive() {
		return active;
	}

	public String getInactiveMessage() {
		return inactiveMessage;
	}

	public void activate() {
		try {
			try {
				BzrPropertiesLoader.loadBzrLibProperties();
				BazaarClientFactory.setupBestAvailableBackend(true);
			} catch (BazaarClientException bce) {
				throw new InfrastructureException(
						"Failure configuring bzr backend", bce);
			}
			repository = BazaarClientFactory.createClient(BazaarClientFactory
					.getPreferredClientType());
			repository.setWorkDir(File.listRoots()[0]);
			active = true;
		} catch (BazaarClientException e) {
			log.error("Connection to Bazaar branch " + getRoot() + " failed: "
					+ e, e);
			// We don't want to throw an exception here because then the system
			// won't start if the repo is down
			// or there is something wrong with the configuration. We also still
			// want this repository to show up
			// in our configuration so the user has a chance to fix the problem.
			active = false;
			inactiveMessage = e.getMessage();
		} catch (InfrastructureException ie) {
			log.error(ie.getMessage(), ie);
			active = false;
			inactiveMessage = ie.getMessage();
		}
	}

	private void deactivate(String message) {
		if (repository != null) {
			repository = null;
		}
		active = false;
		inactiveMessage = message;
	}

	public ViewLinkFormat getViewLinkFormat() {
		if (!isViewLinkSet) {
			final String type = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_TYPE);
			final String linkPathFormat = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_PATH_KEY);
			final String changesetFormat = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_CHANGESET);
			final String fileAddedFormat = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_ADDED);
			final String fileModifiedFormat = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_MODIFIED);
			final String fileReplacedFormat = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_REPLACED);
			final String fileDeletedFormat = properties
					.getString(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_DELETED);

			if (linkPathFormat != null || changesetFormat != null
					|| fileAddedFormat != null || fileModifiedFormat != null
					|| fileReplacedFormat != null || fileDeletedFormat != null)
				viewLinkFormat = new ViewLinkFormat(type, changesetFormat,
						fileAddedFormat, fileModifiedFormat,
						fileReplacedFormat, fileDeletedFormat, linkPathFormat);
			else
				viewLinkFormat = null; /*
										 * [SVN-190] This could happen if the
										 * user clears all the fields in the
										 * Bazaar repository web link
										 * configuration
										 */
			isViewLinkSet = true;
		}

		return viewLinkFormat;
	}

	public BazaarLinkRenderer getLinkRenderer() {
		return linkRenderer;
	}

	protected static String decryptPassword(String encrypted)
			throws IOException {
		if (encrypted == null)
			return null;

		BASE64Decoder decoder = new BASE64Decoder();
		byte[] result = decoder.decodeBuffer(encrypted);

		return new String(result, 0, result.length);
	}

	protected static String encryptPassword(String password) {
		if (password == null)
			return null;

		BASE64Encoder encoder = new BASE64Encoder();
		return encoder.encode(password.getBytes());
	}
}
