package com.atlassian.jira.plugin.ext.bazaar;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.vcs.bazaar.client.BazaarClientPreferences;
import org.vcs.bazaar.client.BazaarPreference;

import com.atlassian.core.util.ClassLoaderUtils;
import com.atlassian.jira.InfrastructureException;

/**
 * exists to load BazaarManagers the old way so that the
 * MultipleBazaarRepositoryManagerImpl doesn't get krufted up with a bunch of
 * legacy code
 * @author Guillermo Gonzalez <guillo.gonzo@gmail.com>
 */
public class BzrPropertiesLoader {

	private static Logger log = Logger.getLogger(BzrPropertiesLoader.class);

	public static final String PROPERTIES_FILE_NAME = "bzr-jira-plugin.properties";

	public static List<BazaarProperties> getBzrProperties()
			throws InfrastructureException {
		Properties allProps = System.getProperties();

		try {
			allProps.load(ClassLoaderUtils.getResourceAsStream(
					PROPERTIES_FILE_NAME,
					MultipleBazaarRepositoryManagerImpl.class));
		} catch (IOException e) {
			throw new InfrastructureException("Problem loading "
					+ PROPERTIES_FILE_NAME + ".", e);
		}

		List<BazaarProperties> propertyList = new ArrayList<BazaarProperties>();
		BazaarProperties defaultProps = getBazaarProperty(-1, allProps);
		if (defaultProps != null) {
			propertyList.add(defaultProps);
		} else {
			log.error("Could not load properties from " + PROPERTIES_FILE_NAME);
			throw new InfrastructureException("Could not load properties from "
					+ PROPERTIES_FILE_NAME);
		}
		BazaarProperties prop;
		int i = 1;
		do {
			prop = getBazaarProperty(i, allProps);
			i++;
			if (prop != null) {
				prop.fillPropertiesFromOther(defaultProps);
				propertyList.add(prop);
			}
		} while (prop != null);

		return propertyList;
	}

	protected static BazaarProperties getBazaarProperty(int index,
			Properties props) {
		String indexStr = "." + Integer.toString(index);
		if (index == -1) {
			indexStr = "";
		}

		if (props
				.containsKey(MultipleBazaarRepositoryManager.BZR_BRANCH_ROOT_KEY
						+ indexStr)) {
			final String bzrRootStr = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_BRANCH_ROOT_KEY
							+ indexStr);
			final String displayName = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_REPOSITORY_NAME
							+ indexStr);

			final String changesetFormat = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_CHANGESET
							+ indexStr);
			final String fileAddedFormat = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_ADDED
							+ indexStr);
			final String fileModifiedFormat = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_MODIFIED
							+ indexStr);
			final String fileReplacedFormat = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_REPLACED
							+ indexStr);
			final String fileDeletedFormat = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_LINKFORMAT_FILE_DELETED
							+ indexStr);

			final String username = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_USERNAME_KEY
							+ indexStr);
			final String password = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_PASSWORD_KEY
							+ indexStr);
			final String privateKeyFile = props
					.getProperty(MultipleBazaarRepositoryManager.BZR_PRIVATE_KEY_FILE
							+ indexStr);
			Boolean revisionIndexing = null;
			if (props
					.containsKey(MultipleBazaarRepositoryManager.BZR_REVISION_INDEXING_KEY
							+ indexStr)) {
				revisionIndexing = Boolean
						.valueOf("true"
								.equalsIgnoreCase(props
										.getProperty(MultipleBazaarRepositoryManager.BZR_REVISION_INDEXING_KEY
												+ indexStr)));
			}
			Integer revisionCacheSize = null;
			if (props
					.containsKey(MultipleBazaarRepositoryManager.BZR_REVISION_CACHE_SIZE_KEY
							+ indexStr)) {
				revisionCacheSize = new Integer(
						props
								.getProperty(MultipleBazaarRepositoryManager.BZR_REVISION_CACHE_SIZE_KEY
										+ indexStr));
			}

			return new BazaarProperties().setRoot(bzrRootStr).setDisplayName(
					displayName).setChangeSetFormat(changesetFormat)
					.setFileAddedFormat(fileAddedFormat)
					.setFileModifiedFormat(fileModifiedFormat)
					.setFileReplacedFormat(fileReplacedFormat)
					.setFileDeletedFormat(fileDeletedFormat)
					.setRevisionIndexing(revisionIndexing)
					.setRevisioningCacheSize(revisionCacheSize);

		} else {
			log.info("No "
					+ MultipleBazaarRepositoryManager.BZR_BRANCH_ROOT_KEY
					+ indexStr + " specified in " + PROPERTIES_FILE_NAME);
			return null;
		}
	}

	public static void loadBzrLibProperties() {
		Properties props = new Properties();
		try {
			props.load(ClassLoaderUtils.getResourceAsStream(
					BzrPropertiesLoader.PROPERTIES_FILE_NAME,
					MultipleBazaarRepositoryManagerImpl.class));
		} catch (IOException e) {
			throw new InfrastructureException(
					"Failure configuring bzr backend, bzr client preferences not found",
					e);
		}
		// final String home =
		// (String)props.get(BazaarPreference.BZR_HOME.name());
		// final String pluginPath =
		// (String)props.get(MultipleBazaarRepositoryManager.BZR_XMLRPC_PORT);
		final String port = (String) props
				.get(MultipleBazaarRepositoryManager.BZR_XMLRPC_PORT);
		final String executable = (String) props
				.get(MultipleBazaarRepositoryManager.BZR_EXECUTABLE);
		BazaarClientPreferences prefs = BazaarClientPreferences.getInstance();
		// prefs.set(BazaarPreference.BZR_HOME, home);
		// prefs.set(BazaarPreference.BZR_PLUGIN_PATH, pluginPath);
		prefs.set(BazaarPreference.BZR_XMLRPC_PORT, port);
		prefs.set(BazaarPreference.EXECUTABLE, executable);
	}
}
