package com.atlassian.jira.plugin.ext.bazaar;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.vcs.bazaar.client.BazaarClientFactory;
import org.vcs.bazaar.client.core.BazaarClientException;

import com.atlassian.jira.InfrastructureException;
import com.atlassian.jira.config.util.IndexPathManager;
import com.atlassian.jira.config.properties.ApplicationProperties;
import com.atlassian.jira.issue.IssueManager;
import com.atlassian.jira.issue.changehistory.ChangeHistoryManager;
import com.atlassian.jira.plugin.ext.bazaar.revisions.RevisionIndexService;
import com.atlassian.jira.plugin.ext.bazaar.revisions.RevisionIndexer;
import com.atlassian.jira.project.version.VersionManager;
import com.atlassian.jira.propertyset.JiraPropertySetFactory;
import com.atlassian.jira.security.PermissionManager;
import com.opensymphony.module.propertyset.PropertySet;

/**
 * This is a wrapper class for many BazaarManagers. Configured via
 * {@link BzrPropertiesLoader#PROPERTIES_FILE_NAME}
 * 
 * @author Dylan Etkin
 * @author Guillermo Gonzalez <guillo.gonzo@gmail.com>
 * @see BazaarManager
 */
public class MultipleBazaarRepositoryManagerImpl implements
		MultipleBazaarRepositoryManager {
	private static Logger log = Logger
			.getLogger(MultipleBazaarRepositoryManagerImpl.class);

	public static final String APP_PROPERTY_PREFIX = "jira.plugins.bazaar";
	public static final String REPO_PROPERTY = "jira.plugins.bazaar.repo";

	public static final String LAST_REPO_ID = "last.repo.id";

	public static final long FIRST_REPO_ID = 1;

	private PropertySet pluginProperties;

	/**
	 * long -> BazaarManager
	 */
	private Map<Long, BazaarManager> managerMap = new HashMap<Long, BazaarManager>();
	private RevisionIndexer revisionIndexer;
	private JiraPropertySetFactory jiraPropertySetFactory;

	private long lastRepoId;

	public MultipleBazaarRepositoryManagerImpl(
			ApplicationProperties applicationProperties,
			VersionManager versionManager, IssueManager issueManager,
			PermissionManager permissionManager,
			ChangeHistoryManager changeHistoryManager,
			JiraPropertySetFactory jiraPropertySetFactory,
			IndexPathManager indexPathManager)

	{
		this.jiraPropertySetFactory = jiraPropertySetFactory;
		// Initialize the SVN tools
		try {
			BzrPropertiesLoader.loadBzrLibProperties();
			BazaarClientFactory.setupBestAvailableBackend(true);
		} catch (BazaarClientException bce) {
			throw new InfrastructureException(
					"Failure configuring bzr backend", bce);
		}
		managerMap = loadManagers();

		boolean revisionIndexing = false;
		for (BazaarManager mgr : managerMap.values()) {
			if (mgr.isRevisionIndexing()) {
				revisionIndexing = true;
				break;
			}
		}

		if (!revisionIndexing) // they might have removed the property - let's
								// check there is no service anyway
		{
			try {
				RevisionIndexService.remove();
			} catch (Exception e) {
				throw new InfrastructureException(
						"Failure removing revision indexing service", e);
			}
		} else {
			// create revision indexer once we know we have succeed initializing
			// our repositories
			revisionIndexer = new RevisionIndexer(this, applicationProperties,
					versionManager, issueManager, permissionManager,
					changeHistoryManager, indexPathManager);
		}
	}

	/**
	 * loads a map of long id to BazaarManager from persistent storage or if
	 * that doesn't exist, system properties
	 */
	protected Map<Long, BazaarManager> loadManagers() {

		Map<Long, BazaarManager> managers = loadManagersFromJiraProperties();

		if (managers.isEmpty()) {
			log
					.info("Could not find any bazaar repositories configured, trying to load from System Properties.");
			try {
				// Try loading from the properties file.
				managers = loadFromProperties();
			} catch (Throwable t) {
				log.error("Could not load properties from "
						+ BzrPropertiesLoader.PROPERTIES_FILE_NAME, t);
				throw new InfrastructureException(
						"Could not load properties from "
								+ BzrPropertiesLoader.PROPERTIES_FILE_NAME, t);
			}
		}

		return managers;
	}

	/**
	 * The bazaar configuration properties are stored in the application
	 * properties. It's not the best place to store collections of information,
	 * like multiple repositories, but it will work. Keys for the properties
	 * look like:
	 * <p/>
	 * jira.plugins.bazaar.<repoId>;<property name>
	 * <p/>
	 * Using this scheme we can get all the properties and put them into buckets
	 * corresponding to the repoId. Then when we have all the properties we can
	 * go about building the BazaarProperties objects and creating our
	 * BazaarRepositoryManagers.
	 */
	private Map<Long, BazaarManager> loadManagersFromJiraProperties() {

		pluginProperties = jiraPropertySetFactory
				.buildCachingDefaultPropertySet(APP_PROPERTY_PREFIX, true);

		lastRepoId = pluginProperties.getLong(LAST_REPO_ID);

		// create the BazaarManagers
		Map<Long, BazaarManager> managers = new LinkedHashMap<Long, BazaarManager>();
		for (long i = FIRST_REPO_ID; i <= lastRepoId; i++) {
			BazaarManager mgr = createManagerFromPropertySet(i,
					jiraPropertySetFactory.buildCachingPropertySet(
							REPO_PROPERTY, new Long(i), true));
			if (mgr != null)
				managers.put(new Long(i), mgr);
		}
		return managers;
	}

	protected BazaarManager createManagerFromPropertySet(long index,
			PropertySet properties) {
		try {
			if (properties.getKeys().size() == 0)
				return null;

			return new BazaarManagerImpl(index, properties);
		} catch (IllegalArgumentException e) {
			log
					.error(
							"Error creating BazaarManager "
									+ index
									+ ". Probably was missing a required field (e.g., repository name or root). Skipping it.",
							e);
			return null;
		}
	}

	/**
	 * loads BazaarManagers from SystemProperties the legacy way
	 */
	protected Map<Long, BazaarManager> loadFromProperties() {
		Map<Long, BazaarManager> managers = new HashMap<Long, BazaarManager>();
		try {
			List<BazaarProperties> properties = BzrPropertiesLoader
					.getBzrProperties();
			for (BazaarProperties property : properties) {
				try {
					BazaarManager mgr = createRepository(property);
					managers.put(new Long(mgr.getId()), mgr);
				} catch (Throwable t) {
					log.warn("Problem adding a bazaar manager", t);
				}
			}
		} catch (Exception e) {
			log.warn("Problem adding a bazaar manager", e);
		}

		return managers;
	}

	public BazaarManager createRepository(BzrProperties properties) {
		long repoId;
		synchronized (this) {
			repoId = ++lastRepoId;
			pluginProperties.setLong(LAST_REPO_ID, lastRepoId);
		}

		PropertySet set = jiraPropertySetFactory.buildCachingPropertySet(
				REPO_PROPERTY, new Long(repoId), true);
		BazaarManager bazaarManager = new BazaarManagerImpl(repoId,
				BzrProperties.Util.fillPropertySet(properties, set));

		managerMap.put(new Long(bazaarManager.getId()), bazaarManager);
		if (isIndexingRevisions()) {
			revisionIndexer.addRepository(bazaarManager);
		}

		return bazaarManager;
	}

	public BazaarManager updateRepository(long repoId, BzrProperties properties) {
		BazaarManager bazaarManager = getRepository(repoId);
		bazaarManager.update(properties);
		return bazaarManager;
	}

	public void removeRepository(long repoId) {
		BazaarManager original = (BazaarManager) managerMap
				.get(new Long(repoId));
		if (original == null) {
			return;
		}
		try {
			managerMap.remove(new Long(repoId));
			for (String key : new ArrayList<String>(original.getProperties()
					.getKeys())) {
				original.getProperties().remove(key);
			}

			if (revisionIndexer != null) {
				revisionIndexer.removeEntries(original);
			}
		} catch (Exception e) {
			throw new InfrastructureException(
					"Could not remove repository index", e);
		}
	}

	public boolean isIndexingRevisions() {
		return revisionIndexer != null;
	}

	public RevisionIndexer getRevisionIndexer() {
		return revisionIndexer;
	}

	public Collection<BazaarManager> getRepositoryList() {
		return managerMap.values();
	}

	public BazaarManager getRepository(long id) {
		return (BazaarManager) managerMap.get(new Long(id));
	}

	public void start() throws Exception {
		if (isIndexingRevisions()) {
			getRevisionIndexer().start();
		}
	}
}
