package com.atlassian.jira.plugin.ext.bazaar.issuetabpanels.changes;

import java.sql.Timestamp;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.vcs.bazaar.client.BazaarStatus;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.IBazaarStatus;
import org.vcs.bazaar.client.commandline.CommandLineLogMessage;

import com.atlassian.jira.plugin.ext.bazaar.MultipleBazaarRepositoryManager;
import com.atlassian.jira.plugin.ext.bazaar.linkrenderer.BazaarLinkRenderer;
import com.atlassian.jira.plugin.issuetabpanel.AbstractIssueAction;
import com.atlassian.jira.plugin.issuetabpanel.IssueTabPanelModuleDescriptor;
import com.atlassian.jira.util.JiraKeyUtils;

/**
 * One item in the 'Bazaar Commits' tab.
 */
public class BazaarRevisionAction extends AbstractIssueAction {

	private final IBazaarLogMessage revision;
	private final long repoId;
	protected final IssueTabPanelModuleDescriptor descriptor;
	protected MultipleBazaarRepositoryManager multipleBazaarRepositoryManager;
	protected Timestamp timePerformed;
	private List<IBazaarStatus> files;

	public BazaarRevisionAction(IBazaarLogMessage logEntry,
			MultipleBazaarRepositoryManager multipleBazaarRepositoryManager,
			IssueTabPanelModuleDescriptor descriptor, long repoId) {
		super(descriptor);
		this.multipleBazaarRepositoryManager = multipleBazaarRepositoryManager;
		this.descriptor = descriptor;
		/* SVN-93 */
		this.revision = new CommandLineLogMessage(logEntry.getRevision()
				.getValue(), logEntry.getCommiter(), logEntry.getAuthor(),
				logEntry.getBranchNick(), logEntry.getTimeStamp(),
				rewriteLogMessage(logEntry.getMessage()), 
				logEntry.getAffectedFiles(),
				logEntry.getMerged(), 
				logEntry.getRevisionId(), 
				logEntry.getParents());
		this.timePerformed = new Timestamp(revision.getDate().getTime());
		this.files = logEntry.getAffectedFiles();
		this.repoId = repoId;
	}

	public List<IBazaarStatus> getFiles() {
		return files;
	}

	protected void populateVelocityParams(Map params) {
		params.put("bzr", this);
	}

	public BazaarLinkRenderer getLinkRenderer() {
		return multipleBazaarRepositoryManager.getRepository(repoId)
				.getLinkRenderer();
	}

	public String getRepositoryDisplayName() {
		return multipleBazaarRepositoryManager.getRepository(repoId)
				.getDisplayName();
	}

	public Date getTimePerformed() {
		return timePerformed;
	}

	public long getRepoId() {
		return repoId;
	}

	public String getUsername() {
		return revision.getAuthor();
	}

	public IBazaarLogMessage getRevision() {
		return revision;
	}

	public boolean isAdded(BazaarStatus logEntryPath) {
		return logEntryPath.contains(BazaarStatusKind.CREATED);
	}

	public boolean isModified(BazaarStatus logEntryPath) {
		return logEntryPath.contains(BazaarStatusKind.MODIFIED);
	}

	public boolean isRenamed(BazaarStatus logEntryPath) {
		return logEntryPath.contains(BazaarStatusKind.RENAMED);
	}

	public boolean isDeleted(BazaarStatus logEntryPath) {
		return logEntryPath.contains(BazaarStatusKind.DELETED);
	}

	public boolean isCopyPath(BazaarStatus status) {
		return status.getPreviousPath() != null
				&& !"".equals(status.getPreviousPath());
	}

	/**
	 * Converts all lower case JIRA issue keys to upper case so that they can be
	 * correctly rendered in the Velocity macro, makelinkedhtml.
	 * 
	 * @param logMessageToBeRewritten
	 *            The SVN log message to be rewritten.
	 * @return The rewritten SVN log message.
	 * @see <a href="http://jira.atlassian.com/browse/SVN-93">SVN-93</a>
	 */
	protected String rewriteLogMessage(final String logMessageToBeRewritten) {
		String logMessage = logMessageToBeRewritten;
		final String logMessageUpperCase = StringUtils.upperCase(logMessage);
		final Set<String> issueKeys = new HashSet<String>(JiraKeyUtils
				.getIssueKeysFromString(logMessageUpperCase));
		for (String issueKey : issueKeys) {
			int indexOfIssueKey;
			int lastIndexOfIssueKey = 0;
			while (lastIndexOfIssueKey < logMessageUpperCase.length()
					&& -1 != (indexOfIssueKey = logMessageUpperCase.indexOf(
							issueKey, lastIndexOfIssueKey))) {
				logMessage = logMessage.replaceFirst(logMessage.substring(
						indexOfIssueKey, indexOfIssueKey + issueKey.length()),
						issueKey);
				lastIndexOfIssueKey = indexOfIssueKey + issueKey.length();
			}
		}

		return logMessage;
	}
}
