package com.atlassian.jira.plugin.ext.bazaar.linkrenderer;

import org.apache.log4j.Logger;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.BazaarStatus;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.IBazaarLogMessage;

import com.atlassian.core.util.StringUtils;
import com.atlassian.jira.plugin.ext.bazaar.BazaarManager;
import com.atlassian.jira.plugin.ext.bazaar.ViewLinkFormat;

/**
 * A link renderer implementation which lets the user specify the format in the
 * properties file, to accommodate various formats (ViewCVS, Fisheye, etc) out
 * there.
 * 
 * @author Chenggong Lu
 * @author Jeff Turner
 * @author Guillermo Gonzalez
 */
public class LinkFormatRenderer implements BazaarLinkRenderer {
	private static Logger log = Logger.getLogger(LinkFormatRenderer.class);
	private String pathLinkFormat;
	private String fileReplacedFormat;
	private String fileAddedFormat;
	private String fileModifiedFormat;
	private String fileDeletedFormat;
	private String changesetFormat;

	public LinkFormatRenderer(BazaarManager bazaarManager) {

		ViewLinkFormat linkFormat = bazaarManager.getViewLinkFormat();
		if (linkFormat != null) {
			if (linkFormat.getChangesetFormat() != null
					&& linkFormat.getChangesetFormat().trim().length() != 0) {
				changesetFormat = linkFormat.getChangesetFormat();
			}

			if (linkFormat.getFileAddedFormat() != null
					&& linkFormat.getFileAddedFormat().trim().length() != 0) {
				fileAddedFormat = linkFormat.getFileAddedFormat();
			}

			if (linkFormat.getFileModifiedFormat() != null
					&& linkFormat.getFileModifiedFormat().trim().length() != 0) {
				fileModifiedFormat = linkFormat.getFileModifiedFormat();
			}

			if (linkFormat.getFileReplacedFormat() != null
					&& linkFormat.getFileReplacedFormat().trim().length() != 0) {
				fileReplacedFormat = linkFormat.getFileReplacedFormat();
			}

			if (linkFormat.getFileDeletedFormat() != null
					&& linkFormat.getFileDeletedFormat().trim().length() != 0) {
				fileDeletedFormat = linkFormat.getFileDeletedFormat();
			}

			if (linkFormat.getViewFormat() != null
					&& linkFormat.getViewFormat().trim().length() != 0) {
				pathLinkFormat = linkFormat.getViewFormat();
			}

		} else {
			log.warn("viewLinkFormat is null");
		}
	}

	public String getCopySrcLink(IBazaarLogMessage revision,
			BazaarStatus logEntryPath) {
		BazaarRevision revisionNumber = revision.getRevision();
		return linkPath(pathLinkFormat, logEntryPath.getPreviousPath(),
				revisionNumber);
	}

	public String getRevisionLink(IBazaarLogMessage revision) {
		return getRevisionLink(revision.getRevision());
	}

	public String getChangePathLink(IBazaarLogMessage revision,
			BazaarStatus logEntryPath) {
		String path = logEntryPath.getPath();
		BazaarRevision revisionNumber = revision.getRevision();

		if (logEntryPath.contains(BazaarStatusKind.MODIFIED)) {
			return linkPath(fileModifiedFormat, path, revisionNumber);
		} else if (logEntryPath.contains(BazaarStatusKind.CREATED)) {
			return linkPath(fileAddedFormat, path, revisionNumber);
		} else if (logEntryPath.contains(BazaarStatusKind.RENAMED)) {
			return linkPath(fileReplacedFormat, path, revisionNumber);
		} else if (logEntryPath.contains(BazaarStatusKind.DELETED)) {
			return linkPath(fileDeletedFormat, path, revisionNumber);
		} else {
			return linkPath(fileReplacedFormat, path, revisionNumber);
		}
	}

	protected String getRevisionLink(BazaarRevision revision) {
		if (changesetFormat != null) {
			try {
				String href = StringUtils.replaceAll(changesetFormat, "${rev}",
						"" + revision.getValue());
				return "<a href=\"" + href + "\">#" + revision.getValue()
						+ "</a>";
			} catch (Exception ex) {
				log.error("format error: " + ex.getMessage(), ex);
			}
		}
		return "#" + revision.getValue();

	}

	private String linkPath(final String format, String path,
			BazaarRevision revision) {

		if (path != null && path.length() > 0 && path.charAt(0) != '/') {
			path = "/" + path;
		}

		if (format != null) {

			try {
				String href = format;
				if (path != null) {
					href = StringUtils.replaceAll(href, "${path}", path);
				}
				href = StringUtils.replaceAll(href, "${rev}", ""
						+ revision.getValue());
				// TODO: use the list of logs to get the previous one
				// FIXME: for the moment the same revno is shown
				href = StringUtils.replaceAll(href, "${rev-1}", ""
						+ (revision.getValue()));

				return "<a href=\"" + href + "\">" + path + "</a>";
			} catch (Exception ex) {
				log.error("format error: " + ex.getMessage(), ex);
			}
		}
		return path;
	}
}
