package com.atlassian.jira.plugin.ext.bazaar;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

import org.jmock.Mock;
import org.jmock.MockObjectTestCase;

import com.atlassian.jira.config.properties.ApplicationProperties;
import com.atlassian.jira.config.util.IndexPathManager;
import com.atlassian.jira.issue.IssueManager;
import com.atlassian.jira.issue.changehistory.ChangeHistoryManager;
import com.atlassian.jira.plugin.ext.bazaar.revisions.RevisionIndexService;
import com.atlassian.jira.project.version.VersionManager;
import com.atlassian.jira.propertyset.JiraPropertySetFactory;
import com.atlassian.jira.security.PermissionManager;
import com.atlassian.jira.service.JiraServiceContainer;
import com.atlassian.jira.service.ServiceManager;
import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.module.propertyset.memory.MemoryPropertySet;

public class TestMultipleBazaarRepositoryManagerImpl extends MockObjectTestCase {

    private Mock mockApplicationProperties;

    private ApplicationProperties applicationProperties;

    private Mock mockVersionManager;

    private VersionManager versionManager;

    private Mock mockIssueManager;

    private IssueManager issueManager;

    private Mock mockPermissionManager;

    private PermissionManager permissionManager;

    private Mock mockIndexPathManager;

    private IndexPathManager indexPathManager;

    private ChangeHistoryManager changeHistoryManager;

    private Mock mockChangeHistoryManager;

    private Mock mockJiraPropertySetFactory;

    private JiraPropertySetFactory jiraPropertySetFactory;

    private PropertySet jiraPropertySet;

    private Mock mockServiceManager;

    private ServiceManager serviceManager;

    protected void setUp() throws Exception {

        mockApplicationProperties = new Mock(ApplicationProperties.class);
        applicationProperties = (ApplicationProperties) mockApplicationProperties.proxy();

        mockVersionManager = new Mock(VersionManager.class);
        versionManager = (VersionManager) mockVersionManager.proxy();

        mockIssueManager = new Mock(IssueManager.class);
        issueManager = (IssueManager) mockIssueManager.proxy();

        mockPermissionManager = new Mock(PermissionManager.class);
        permissionManager = (PermissionManager) mockPermissionManager.proxy();

        mockIndexPathManager = new Mock(IndexPathManager.class);
        indexPathManager = (IndexPathManager) mockIndexPathManager.proxy();

        mockChangeHistoryManager = new Mock(ChangeHistoryManager.class);
        changeHistoryManager = (ChangeHistoryManager) mockChangeHistoryManager.proxy();

        mockJiraPropertySetFactory = new Mock(JiraPropertySetFactory.class);
        jiraPropertySetFactory = (JiraPropertySetFactory) mockJiraPropertySetFactory.proxy();

        jiraPropertySet = new MemoryPropertySet();
        jiraPropertySet.init(new HashMap(), new HashMap());

        jiraPropertySet.setLong(MultipleBazaarRepositoryManagerImpl.LAST_REPO_ID, 2);
        jiraPropertySet.setString(MultipleBazaarRepositoryManager.BZR_BRANCH_ROOT_KEY, System.getProperty("bzr.root"));

        mockServiceManager = new Mock(ServiceManager.class);
        serviceManager = (ServiceManager) mockServiceManager.proxy();
        RevisionIndexService.setServiceManager(serviceManager);
    }

    protected void tearDown() throws Exception {
        RevisionIndexService.setServiceManager(null);
        super.tearDown();
    }

    protected MultipleBazaarRepositoryManager getMultipleBazaarRepositoryManager() {
        return new MultipleBazaarRepositoryManagerImpl(
		applicationProperties, versionManager, issueManager, permissionManager, changeHistoryManager, jiraPropertySetFactory, indexPathManager
		
        );
    }

    public void testInstantiateMultipleBazaarRepositoryManagerWithNoIndexingRepositoryFromJiraProperties() {
         /* Called by MultipleBazaarRepositoryManagerImpl#loadManagersFromJiraProperties */
        mockJiraPropertySetFactory
                .expects(once())
                .method("buildCachingDefaultPropertySet")
                .with(eq(MultipleBazaarRepositoryManagerImpl.APP_PROPERTY_PREFIX), eq(true))
                .will(returnValue(jiraPropertySet));
        /* To return a set of properties to be used in the creation of a BazaarManager. We will be creating two repos */
        mockJiraPropertySetFactory
                .expects(exactly(2))
                .method("buildCachingPropertySet")
                .with(eq(MultipleBazaarRepositoryManagerImpl.REPO_PROPERTY), isA(Long.class), eq(true))
                .will(returnValue(jiraPropertySet));

        /* These methods will be called when there are no indexing repositories */
        mockServiceManager
                .expects(once())
                .method("getServiceWithName")
                .with(eq(RevisionIndexService.REVISION_INDEX_SERVICE_NAME))
                .will(returnValue(new Mock(JiraServiceContainer.class).proxy()));
        mockServiceManager
                .expects(once())
                .method("removeServiceByName")
                .with(eq(RevisionIndexService.REVISION_INDEX_SERVICE_NAME));

        /* No indexing repo */
        jiraPropertySet.setBoolean(MultipleBazaarRepositoryManager.BZR_REVISION_INDEXING_KEY, false);

        MultipleBazaarRepositoryManager multipleBazaarRepositoryManager = getMultipleBazaarRepositoryManager();
        Collection subversionManagers = multipleBazaarRepositoryManager.getRepositoryList();
        BazaarManager bzrMgr;
        Iterator bzrMgrIter;

        assertEquals(2, subversionManagers.size());

        bzrMgrIter = subversionManagers.iterator();
        for (int i = 1; bzrMgrIter.hasNext(); ++i) {
            bzrMgr = (BazaarManager) bzrMgrIter.next();
            assertEquals(i, bzrMgr.getId());
        }

        assertNull(multipleBazaarRepositoryManager.getRevisionIndexer());
        mockServiceManager.verify();
    }

    public void testInstantiateMultipleBazaarRepositoryManagerWithIndexingRepositoriesFromJiraProperties() {
         /* Called by MultipleBazaarRepositoryManagerImpl#loadManagersFromJiraProperties */
        mockJiraPropertySetFactory
                .expects(once())
                .method("buildCachingDefaultPropertySet")
                .with(eq(MultipleBazaarRepositoryManagerImpl.APP_PROPERTY_PREFIX), eq(true))
                .will(returnValue(jiraPropertySet));
        /* To return a set of properties to be used in the creation of a BazaarManager. We will be creating two repos */
        mockJiraPropertySetFactory
                .expects(exactly(2))
                .method("buildCachingPropertySet")
                .with(eq(MultipleBazaarRepositoryManagerImpl.REPO_PROPERTY), isA(Long.class), eq(true))
                .will(returnValue(jiraPropertySet));

        /* An indexing repo */
        jiraPropertySet.setBoolean(MultipleBazaarRepositoryManager.BZR_REVISION_INDEXING_KEY, true);

        MultipleBazaarRepositoryManager multipleBazaarRepositoryManager = getMultipleBazaarRepositoryManager();
        Collection subversionManagers = multipleBazaarRepositoryManager.getRepositoryList();
        BazaarManager bzrMgr;
        Iterator bzrMgrIter;

        assertEquals(2, subversionManagers.size());

        bzrMgrIter = subversionManagers.iterator();
        for (int i = 1; bzrMgrIter.hasNext(); ++i) {
            bzrMgr = (BazaarManager) bzrMgrIter.next();
            assertEquals(i, bzrMgr.getId());
        }

        assertNotNull(multipleBazaarRepositoryManager.getRevisionIndexer());
    }

    public void testInstantiateMultipleBazaarRepositoryManagerWithIndexingRepositoriesFromProperties() {
        /* For multiple repositories from properties file */
        System.setProperty(MultipleBazaarRepositoryManager.BZR_BRANCH_ROOT_KEY + ".1", "file://foo/bar");
        System.setProperty(MultipleBazaarRepositoryManager.BZR_REPOSITORY_NAME + ".1", "Foobar");

         /* Called by MultipleBazaarRepositoryManagerImpl#loadManagersFromJiraProperties */
        mockJiraPropertySetFactory
                .expects(once())
                .method("buildCachingDefaultPropertySet")
                .with(eq(MultipleBazaarRepositoryManagerImpl.APP_PROPERTY_PREFIX), eq(true))
                .will(returnValue(jiraPropertySet));
        /*
         * This will get called in MultipleBazaarRepositoryManagerImpl#createRepository. We will still return jiraPropertySet because
         * bzrProperties is basically the same as jiraPropertySet after it is filtered by the jiraPropertySetFactory.
         */
        mockJiraPropertySetFactory
                .expects(exactly(2))
                .method("buildCachingPropertySet")
                .with(eq(MultipleBazaarRepositoryManagerImpl.REPO_PROPERTY), isA(Long.class), eq(true))
                .will(returnValue(jiraPropertySet));

        /* Nothing repo defined in application properties */
        final long initialLastRepoId = 0;

        jiraPropertySet.setLong(MultipleBazaarRepositoryManagerImpl.LAST_REPO_ID, initialLastRepoId);
        /* BZR properties */

        MultipleBazaarRepositoryManager multipleBazaarRepositoryManager = getMultipleBazaarRepositoryManager();
        Collection subversionManagers = multipleBazaarRepositoryManager.getRepositoryList();
        
        assertEquals(2, subversionManagers.size());
        assertNotNull(multipleBazaarRepositoryManager.getRevisionIndexer());
    }
}
