package com.atlassian.jira.plugin.ext.bazaar.issuetabpanels.changes;

import java.util.Collections;
import java.util.Date;
import java.util.List;

import org.jmock.Mock;
import org.jmock.cglib.MockObjectTestCase;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.commandline.CommandLineLogMessage;

import com.atlassian.core.util.map.EasyMap;
import com.atlassian.jira.config.properties.ApplicationProperties;
import com.atlassian.jira.config.util.IndexPathManager;
import com.atlassian.jira.issue.Issue;
import com.atlassian.jira.issue.IssueManager;
import com.atlassian.jira.issue.changehistory.ChangeHistoryManager;
import com.atlassian.jira.issue.tabpanels.GenericMessageAction;
import com.atlassian.jira.plugin.ext.bazaar.MultipleBazaarRepositoryManager;
import com.atlassian.jira.plugin.ext.bazaar.Util;
import com.atlassian.jira.plugin.ext.bazaar.revisions.RevisionIndexer;
import com.atlassian.jira.project.version.VersionManager;
import com.atlassian.jira.security.PermissionManager;
import com.atlassian.jira.security.Permissions;
import com.atlassian.jira.util.EasyList;

public class TestBazaarRevisionsTabPanel extends MockObjectTestCase {

    private Mock mockMultipleBazaarRepositoryManager;

    private Mock mockRevisionIndexer;

    private Mock mockPermissionManager;

    BazaarRevisionsTabPanel getBazaarRevisionsTabPanel() {
        BazaarRevisionsTabPanel bazaarRevisionsTabPanel;

        bazaarRevisionsTabPanel =  new BazaarRevisionsTabPanel(
                (MultipleBazaarRepositoryManager) mockMultipleBazaarRepositoryManager.proxy(),
                (PermissionManager) mockPermissionManager.proxy()) {
            protected String getText(String key) {
                return key;
            }

            protected BazaarRevisionAction createBazaarRevisionAction(long repoId, IBazaarLogMessage logEntry) {
                return new BazaarRevisionAction(logEntry, multipleBazaarRepositoryManager, descriptor, repoId) {
                    protected String rewriteLogMessage(String logMessageToBeRewritten) {
                        return logMessageToBeRewritten;
                    }
                };
            }
        };

        return bazaarRevisionsTabPanel;
    }

    protected void setUp() throws Exception {
        MultipleBazaarRepositoryManager multipleBazaarRepositoryManager;

        super.setUp();

        mockPermissionManager = new Mock(PermissionManager.class);

        mockMultipleBazaarRepositoryManager = new Mock(MultipleBazaarRepositoryManager.class);
        mockMultipleBazaarRepositoryManager.expects(atLeastOnce()).method("getRepositoryList").withNoArguments().will(returnValue(Collections.EMPTY_LIST));
        multipleBazaarRepositoryManager = (MultipleBazaarRepositoryManager) mockMultipleBazaarRepositoryManager.proxy();

        mockRevisionIndexer = mock(
                        TestRevisionIndexer.class,
                        new Class[] {
                                MultipleBazaarRepositoryManager.class,
                                ApplicationProperties.class,
                                VersionManager.class,
                                IssueManager.class,
                                PermissionManager.class,
                                ChangeHistoryManager.class
                        },
                        new Object[] {
                                multipleBazaarRepositoryManager,
                                null,
                                null,
                                null,
                                mockPermissionManager.proxy(),
                                null
                        }
        );
    }

    public void testGetActionsWhenIssueHasNullBzrLogEntries() {
        Mock mockIssue;
        RevisionIndexer revisionIndexer;
        Issue issue;
        BazaarRevisionsTabPanel bazaarRevisionsTabPanel;
        List actions;

        mockIssue = new Mock(Issue.class);
        mockIssue.expects(once()).method("getKey").withNoArguments().will(returnValue("TST-1"));
        issue = (Issue) mockIssue.proxy();

        mockRevisionIndexer.reset();
        mockRevisionIndexer.expects(once()).method("getLogEntriesByRepository").with(same(issue)).will(returnValue(null));
        revisionIndexer = (RevisionIndexer) mockRevisionIndexer.proxy();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRevisionIndexer").withNoArguments().will(returnValue(revisionIndexer));

        bazaarRevisionsTabPanel = getBazaarRevisionsTabPanel();
        actions = bazaarRevisionsTabPanel.getActions(issue, null);

        assertNotNull(actions);
        assertEquals(1, actions.size()); /* List should only have a GenericMessageAction element */
        assertTrue(GenericMessageAction.class.isAssignableFrom(actions.get(0).getClass()));
    }

    public void testGetActionsWhenIssueHasEmptyBzrLogEntries() {
        Mock mockIssue;
        RevisionIndexer revisionIndexer;
        Issue issue;
        BazaarRevisionsTabPanel bazaarRevisionsTabPanel;
        List actions;

        mockIssue = new Mock(Issue.class);
        mockIssue.expects(once()).method("getKey").withNoArguments().will(returnValue("TST-1"));
        issue = (Issue) mockIssue.proxy();

        mockRevisionIndexer.reset();
        mockRevisionIndexer.expects(once()).method("getLogEntriesByRepository").with(same(issue)).will(returnValue(Collections.EMPTY_MAP));
        revisionIndexer = (RevisionIndexer) mockRevisionIndexer.proxy();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRevisionIndexer").withNoArguments().will(returnValue(revisionIndexer));

        bazaarRevisionsTabPanel = getBazaarRevisionsTabPanel();
        actions = bazaarRevisionsTabPanel.getActions(issue, null);

        assertNotNull(actions);
        assertEquals(1, actions.size()); /* List should only have a GenericMessageAction element */
        assertTrue(GenericMessageAction.class.isAssignableFrom(actions.get(0).getClass()));
    }

    public void testGetActionsWhenIssueHasSomeBzrLogEntries() {
        Mock mockIssue;
        RevisionIndexer revisionIndexer;
        Issue issue;
        BazaarRevisionsTabPanel bazaarRevisionsTabPanel;
        List actions;

        mockIssue = new Mock(Issue.class);
        mockIssue.expects(once()).method("getKey").withNoArguments().will(returnValue("TST-1"));
        issue = (Issue) mockIssue.proxy();

        mockRevisionIndexer.reset();
        mockRevisionIndexer.expects(once()).method("getLogEntriesByRepository").with(same(issue)).will(returnValue(
                EasyMap.build(
                        1L, EasyList.build(new CommandLineLogMessage("2", "dchui", "branch_nick", Util.format(new Date()), "foo TST-1", null, null, null, null)),
                        2L, EasyList.build(new CommandLineLogMessage("1", "dchui", "branch_nick", Util.format(new Date()), "bar TST-1", null, null, null, null))
                )
        ));
        revisionIndexer = (RevisionIndexer) mockRevisionIndexer.proxy();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRevisionIndexer").withNoArguments().will(returnValue(revisionIndexer));

        bazaarRevisionsTabPanel = getBazaarRevisionsTabPanel();
        actions = bazaarRevisionsTabPanel.getActions(issue, null);

        assertNotNull(actions);
        assertEquals(2, actions.size()); /* There should be two instances of BazaarRevisionAction returend */
        assertTrue(BazaarRevisionAction.class.isAssignableFrom(actions.get(0).getClass()));
        assertTrue(BazaarRevisionAction.class.isAssignableFrom(actions.get(1).getClass()));
    }

    public void testShowPanelWhenMultipleBazaarManagerIsNotIndexing() {
        BazaarRevisionsTabPanel bazaarRevisionsTabPanel;

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(exactly(2)).method("isIndexingRevisions").withNoArguments().will(returnValue(false));

        mockPermissionManager.reset();
        mockPermissionManager.expects(never()).method("hasPermission").with(eq(Permissions.VIEW_VERSION_CONTROL), NULL, NULL);

        bazaarRevisionsTabPanel = getBazaarRevisionsTabPanel();

        assertFalse(bazaarRevisionsTabPanel.showPanel(null, null)); /* Not indexing and without permissions */
        assertFalse(bazaarRevisionsTabPanel.showPanel(null, null)); /* Not indexing but with permissions */
    }

    public void testShowPanelWhenMultipleBazaarManagerIsIndexing() {
        BazaarRevisionsTabPanel bazaarRevisionsTabPanel;

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(exactly(2)).method("isIndexingRevisions").withNoArguments().will(returnValue(true));

        mockPermissionManager.reset();
        mockPermissionManager.expects(exactly(2)).method("hasPermission").with(eq(Permissions.VIEW_VERSION_CONTROL), NULL, NULL).will(
                onConsecutiveCalls(
                        returnValue(false),
                        returnValue(true)
                )
        );

        bazaarRevisionsTabPanel = getBazaarRevisionsTabPanel();

        assertFalse(bazaarRevisionsTabPanel.showPanel(null, null)); /* Not indexing and without permissions */
        assertTrue(bazaarRevisionsTabPanel.showPanel(null, null)); /* Not indexing but with permissions */
    }

    protected static class TestRevisionIndexer extends RevisionIndexer {

        public TestRevisionIndexer(MultipleBazaarRepositoryManager multipleBazaarRepositoryManager, ApplicationProperties applicationProperties, VersionManager versionManager, IssueManager issueManager, PermissionManager permissionManager, ChangeHistoryManager changeHistoryManager, IndexPathManager indexPathManager) {
            super(multipleBazaarRepositoryManager, applicationProperties, versionManager, issueManager, permissionManager, changeHistoryManager, indexPathManager);
        }
    }
}
