package com.atlassian.jira.plugin.ext.bazaar.revisions;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.SystemUtils;
import org.apache.lucene.document.DateField;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermDocs;
import org.jmock.Mock;
import org.jmock.MockObjectTestCase;
import org.jmock.core.constraint.IsAnything;
import org.ofbiz.core.entity.GenericValue;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.commandline.CommandLineLogMessage;
import org.vcs.bazaar.client.core.BazaarClientException;

import com.atlassian.core.ofbiz.test.mock.MockGenericValue;
import com.atlassian.core.util.collection.EasyList;
import com.atlassian.core.util.map.EasyMap;
import com.atlassian.jira.config.properties.APKeys;
import com.atlassian.jira.config.properties.ApplicationProperties;
import com.atlassian.jira.config.util.IndexPathManager;
import com.atlassian.jira.issue.Issue;
import com.atlassian.jira.issue.IssueManager;
import com.atlassian.jira.issue.MutableIssue;
import com.atlassian.jira.issue.changehistory.ChangeHistoryManager;
import com.atlassian.jira.issue.index.IndexException;
import com.atlassian.jira.plugin.ext.bazaar.BazaarManager;
import com.atlassian.jira.plugin.ext.bazaar.MultipleBazaarRepositoryManager;
import com.atlassian.jira.plugin.ext.bazaar.Util;
import com.atlassian.jira.project.version.Version;
import com.atlassian.jira.project.version.VersionManager;
import com.atlassian.jira.security.PermissionManager;
import com.atlassian.jira.security.Permissions;
import com.atlassian.jira.util.LuceneUtils;
import com.atlassian.plugin.util.FileUtils;

/**
 *
 */
public class TestRevisionIndexer extends MockObjectTestCase {

    private static final String MY_BS_INDEX_PATH = "mybsindexpath";

    private static final String MY_NEW_BS_INDEX_PATH = "mynewbsindexpath";

    private Mock mockMultipleBazaarRepositoryManager;

    private MultipleBazaarRepositoryManager multipleBazaarRepositoryManager;

    private Mock mockIndexPathManager;

    private IndexPathManager indexPathManager;

    private Mock mockApplicationProperties;

    private ApplicationProperties applicationProperties;

    private Mock mockBzrMgr;

    private BazaarManager bzrMgr;

    private Mock mockPermissionManager;

    private PermissionManager permissionManager;

    private ChangeHistoryManager changeHistoryManager;

    private Mock mockChangeHistoryManager;

    private Mock mockIssueManager;

    private IssueManager issueManager;

    private Mock mockVersionManager;

    private VersionManager versionManager;

    private File temporaryIndexDirectory;

    private String getIndexPath() {
        return new StringBuffer(temporaryIndexDirectory.getAbsolutePath()).append(SystemUtils.FILE_SEPARATOR).append(RevisionIndexer.REVISIONS_INDEX_DIRECTORY).toString();
    }

    protected void setUp() throws Exception
    {
        super.setUp();

        /* Setup temporary index directory */
        do {
            temporaryIndexDirectory = new File(SystemUtils.JAVA_IO_TMPDIR, RandomStringUtils.randomAlphanumeric(16));
        } while (temporaryIndexDirectory.exists());
        temporaryIndexDirectory.mkdirs();

        mockApplicationProperties = new Mock(ApplicationProperties.class);
        mockApplicationProperties
                .expects(atLeastOnce())
                .method("getString")
                .with(eq(APKeys.JIRA_PATH_INDEX))
                .will(returnValue(temporaryIndexDirectory.getAbsolutePath()));
        applicationProperties = (ApplicationProperties) mockApplicationProperties.proxy();

        mockMultipleBazaarRepositoryManager = new Mock(MultipleBazaarRepositoryManager.class);
        mockMultipleBazaarRepositoryManager
                .expects(atLeastOnce())
                .method("getRepositoryList")
                .withNoArguments()
                .will(returnValue(Collections.EMPTY_LIST));
        multipleBazaarRepositoryManager = (MultipleBazaarRepositoryManager) mockMultipleBazaarRepositoryManager.proxy();

        mockIndexPathManager = new Mock(IndexPathManager.class);
        indexPathManager = (IndexPathManager) mockIndexPathManager.proxy();

        mockBzrMgr = new Mock(BazaarManager.class);
        bzrMgr = (BazaarManager) mockBzrMgr.proxy();

        mockPermissionManager = new Mock(PermissionManager.class);
        permissionManager = (PermissionManager) mockPermissionManager.proxy();

        List<String> previousIssueKeys = new ArrayList<String>();
        previousIssueKeys.add("APR-137");
        previousIssueKeys.add("FPR-49");
        mockChangeHistoryManager = new Mock(ChangeHistoryManager.class);
        mockChangeHistoryManager
                .expects(atMostOnce())
                .method("getPreviousIssueKeys")
                .will(returnValue(previousIssueKeys));
        changeHistoryManager = (ChangeHistoryManager) mockChangeHistoryManager.proxy();

        mockIssueManager = new Mock(IssueManager.class);
        issueManager = (IssueManager) mockIssueManager.proxy();

        mockVersionManager = new Mock(VersionManager.class);
        versionManager = (VersionManager) mockVersionManager.proxy();
    }

    protected void tearDown() throws Exception {
        if (temporaryIndexDirectory.exists()) {
            FileUtils.deleteDir(temporaryIndexDirectory);
        }
        super.tearDown();
    }

    protected RevisionIndexer getRevisionIndexer() {
        return new RevisionIndexer(
                multipleBazaarRepositoryManager,
                applicationProperties,
                versionManager, issueManager, permissionManager, changeHistoryManager, indexPathManager);
    }

    public void testIndexDirectoryNotCached()
    {
        mockApplicationProperties.reset();
        mockApplicationProperties
                .expects(exactly(2))
                .method("getString")
                .with(eq(APKeys.JIRA_PATH_INDEX))
                .will(returnValue(MY_BS_INDEX_PATH));

        RevisionIndexer revisionIndexer = getRevisionIndexer();
        assertEquals(MY_BS_INDEX_PATH + System.getProperty("file.separator") + RevisionIndexer.REVISIONS_INDEX_DIRECTORY, revisionIndexer.getIndexPath());

        // Update the index path
        mockApplicationProperties.reset();
        mockApplicationProperties
                .expects(exactly(2))
                .method("getString")
                .with(eq(APKeys.JIRA_PATH_INDEX))
                .will(returnValue(MY_NEW_BS_INDEX_PATH));

        assertEquals(MY_NEW_BS_INDEX_PATH + System.getProperty("file.separator") + RevisionIndexer.REVISIONS_INDEX_DIRECTORY, revisionIndexer.getIndexPath());
        assertNotSame(MY_BS_INDEX_PATH + System.getProperty("file.separator") + RevisionIndexer.REVISIONS_INDEX_DIRECTORY, revisionIndexer.getIndexPath());
    }

    public void testUpdateIndexWithNoIndexDirectorWhenEverythingIsOk() throws IOException, IndexException {
        IBazaarLogMessage bzrLogEntry;
        RevisionIndexer revisionIndexer;

        /* Setup the repository to return one bzr entry (revision 0) */
        bzrLogEntry = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "foo TST-1", null, null, null, null);

        mockBzrMgr.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBzrMgr.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
        mockBzrMgr.expects(atLeastOnce()).method("getLogEntries").with(eq(BazaarRevision.INVALID)).will(returnValue(EasyList.build(bzrLogEntry)));
        mockBzrMgr.expects(atLeastOnce()).method("getDisplayName").withNoArguments();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager
                .expects(atLeastOnce())
                .method("getRepositoryList")
                .withNoArguments()
                .will(returnValue(EasyList.build(bzrMgr)));

        revisionIndexer = new RevisionIndexer(multipleBazaarRepositoryManager,
                applicationProperties,
                null, null, null, null, null) {
            protected boolean isKeyInString(IBazaarLogMessage logEntry) { return true; }
            protected List getIssueKeysFromString(IBazaarLogMessage logEntry) { return EasyList.build( "TST-1" ); }
            protected String getProjectKeyFromIssueKey(String issueKey) { return "TST"; }
        };
        revisionIndexer.updateIndex();

        /* Check if a document gets indexed */
        IndexReader indexReader = LuceneUtils.getIndexReader(revisionIndexer.getIndexPath());
        assertEquals(1, indexReader.numDocs());
        indexReader.close();
    }

    public void testUpdateIndexWithNoIndexDirectoryWhenEntryHasEmptyCommitMessage() throws IOException, IndexException {
        IBazaarLogMessage bzrLogEntry;
        RevisionIndexer revisionIndexer;

        /* Setup the repository to return one bzr entry (revision 0) */
        bzrLogEntry = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "", null, null, null, null);

        mockBzrMgr.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBzrMgr.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
        mockBzrMgr.expects(atLeastOnce()).method("getLogEntries").with(eq(BazaarRevision.INVALID)).will(returnValue(EasyList.build(bzrLogEntry)));

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager
                .expects(atLeastOnce())
                .method("getRepositoryList")
                .withNoArguments()
                .will(returnValue(EasyList.build(bzrMgr)));

        revisionIndexer = getRevisionIndexer();
        revisionIndexer.updateIndex();

        /* Check nothing gets indexed */
        IndexReader indexReader = LuceneUtils.getIndexReader(revisionIndexer.getIndexPath());
        assertEquals(0, indexReader.numDocs());
        indexReader.close();
    }

    public void testUpdateIndexWithNoIndexDirectoryWhenEntryCommitMessageHasNoJiraKeys() throws IOException, IndexException {
        IBazaarLogMessage bzrLogEntry;
        RevisionIndexer revisionIndexer;

        /* Setup the repository to return one bzr entry (revision 0) */
        bzrLogEntry = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "There is no JIRA key here.", null, null, null, null);

        mockBzrMgr.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBzrMgr.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
        mockBzrMgr.expects(atLeastOnce()).method("getLogEntries").with(eq(BazaarRevision.INVALID)).will(returnValue(EasyList.build(bzrLogEntry)));
        mockBzrMgr.expects(atLeastOnce()).method("getDisplayName").withNoArguments();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager
                .expects(atLeastOnce())
                .method("getRepositoryList")
                .withNoArguments()
                .will(returnValue(EasyList.build(bzrMgr)));

        revisionIndexer = new RevisionIndexer(multipleBazaarRepositoryManager,
                applicationProperties,
                null, null, null, null, null) {
            protected boolean isKeyInString(IBazaarLogMessage logEntry) { return false; }
        };
        revisionIndexer.updateIndex();

        /* Check nothing gets indexed */
        IndexReader indexReader = LuceneUtils.getIndexReader(revisionIndexer.getIndexPath());
        assertEquals(0, indexReader.numDocs());
        indexReader.close();
    }


    /**
     * Tests that the RevisionIndexer will handle each repository even if method calls on prior repositories throw
     * exceptions.
     *
     * @throws Exception wheneva
     */
    public void testUpdateIndexRepositoryThrowsException() throws Exception
    {
        Mock mockBazaarManagerExploding;
        Mock mockBazaarManagerNonExploding;

        /* The exploding BazaarManager */
        mockBazaarManagerExploding = new Mock(BazaarManager.class);
        mockBazaarManagerExploding.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBazaarManagerExploding.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
        mockBazaarManagerExploding.expects(atLeastOnce()).method("getDisplayName").withNoArguments().will(returnValue("Repository 1"));
        mockBazaarManagerExploding
                .expects(once())
                .method("getLogEntries")
                .withAnyArguments()
                .will(throwException(new RuntimeException("Exploding MockBazaarManager has exploded.")));

        /* The non exploding BazaarManager */
        mockBazaarManagerNonExploding = new Mock(BazaarManager.class);
        mockBazaarManagerNonExploding.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBazaarManagerNonExploding.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
//        mockBazaarManagerNonExploding.expects(atLeastOnce()).method("getDisplayName").withNoArguments().will(returnValue("Repository 2"));
        mockBazaarManagerNonExploding
                .expects(once())
                .method("getLogEntries")
                .withAnyArguments()
                .will(returnValue(Collections.EMPTY_LIST));

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager
                .expects(atLeastOnce())
                .method("getRepositoryList")
                .withNoArguments()
                .will(returnValue(EasyList.build( mockBazaarManagerExploding.proxy(), mockBazaarManagerNonExploding.proxy() )));

        mockApplicationProperties.reset();
        mockApplicationProperties
                .expects(atLeastOnce())
                .method("getString")
                .with(eq(APKeys.JIRA_PATH_INDEX))
                .will(returnValue(temporaryIndexDirectory.getAbsolutePath()));

        // instantiate class under test
        RevisionIndexer ri = new RevisionIndexer(
                (MultipleBazaarRepositoryManager) mockMultipleBazaarRepositoryManager.proxy(),
                (ApplicationProperties) mockApplicationProperties.proxy(),
                null, null, null, null, null) {
            protected BazaarRevision getLatestIndexedRevision(long repoId)
            {
                return BazaarRevision.INVALID;
            }
        };

        try
        {
            ri.updateIndex();
        }
        catch (RuntimeException e)
        {
            fail("The exception thrown from one repository should not propagate out of the all-repo index update method.");
        }
        // from this test point of view we don't really care which methods are called on the repo, if any were called,
        // then the RevisionIndexer didn't catch exceptions outside the while loop
//        assertTrue("The RevisionIndexer probably blew up attempting to index the first repo (SVN-158)", mildManneredRepo.isMethodsCalled());
        mockBazaarManagerNonExploding.verify();
        mockMultipleBazaarRepositoryManager.verify();
    }

    public void testGetLogEntriesByRepositoryWhenIndexDirectoryDoesNotExist() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        RevisionIndexer revisionIndexer;

        mockIssue = new Mock(Issue.class);
        mockIssue.expects(atLeastOnce()).method("getKey").withNoArguments().will(returnValue("TSTS-1"));
        issue = (Issue) mockIssue.proxy();

        mockApplicationProperties.reset();
        mockApplicationProperties
                .expects(atLeastOnce())
                .method("getString")
                .with(eq(APKeys.JIRA_PATH_INDEX))
                .will(returnValue(temporaryIndexDirectory.getAbsolutePath() + "-aPrefixMakingItAnInvalidDirectory"));

        revisionIndexer = getRevisionIndexer();
        assertNull(revisionIndexer.getLogEntriesByRepository(issue));
    }

    public void testGetLogEntriesByRepositoryWhenIssueKeyDoesNotMatch() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        RevisionIndexer revisionIndexer;
        IndexWriter indexWriter;
        Document doc;

        /* Setup the issue to be used in the query */
        mockIssue = new Mock(Issue.class);
        mockIssue.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(new Long(10000L)));
        mockIssue.expects(atLeastOnce()).method("getKey").withNoArguments().will(returnValue("TST-1"));
        issue = (Issue) mockIssue.proxy();

        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-2", Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);
        indexWriter.flush();
        indexWriter.close();

        revisionIndexer = getRevisionIndexer();
        assertEquals(0, revisionIndexer.getLogEntriesByRepository(issue).size());
    }

    public void testGetLogEntriesByRepositoryWhenRevisionCannotBeFoundInRepository() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        RevisionIndexer revisionIndexer;
        IndexWriter indexWriter;
        Document doc;

        /* Setup the issue to be used in the query */
        mockIssue = new Mock(Issue.class);
        mockIssue.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(new Long(10000L)));
        mockIssue.expects(atLeastOnce()).method("getKey").withNoArguments().will(returnValue("TST-1"));
        issue = (Issue) mockIssue.proxy();

        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);
        indexWriter.flush();
        indexWriter.close();

        mockBzrMgr.expects(once()).method("getLogEntry").with(eq(BazaarRevision.getRevision(1))).will(returnValue(null));
        mockMultipleBazaarRepositoryManager
                .expects(once())
                .method("getRepository")
                .with(eq(1L))
                .will(returnValue(bzrMgr));

        revisionIndexer = getRevisionIndexer();
        assertEquals(0, revisionIndexer.getLogEntriesByRepository(issue).size());
    }

    public void testGetLogEntriesByRepositorySorted() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        RevisionIndexer revisionIndexer;
        IndexWriter indexWriter;
        Document doc;

        /* Setup the issue to be used in the query */
        mockIssue = new Mock(Issue.class);
        mockIssue.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(new Long(10000L)));
        mockIssue.expects(atLeastOnce()).method("getKey").withNoArguments().will(returnValue("TST-1"));
        issue = (Issue) mockIssue.proxy();

        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        /* This is the first matching document with revision 1 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        /* This is the second matching document with revision 0 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        indexWriter.flush();
        indexWriter.close();

        IBazaarLogMessage firstCommit = new CommandLineLogMessage("2", "dchui", "branch_nick", new Date().toString(), "First commit for: TST-1", null, null, null, null);
        IBazaarLogMessage secondCommit = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "Second commmit fix: TST-1 ", null, null, null, null);

        mockBzrMgr.expects(exactly(2)).method("getLogEntry").with(isA(BazaarRevision.class)).will(
                onConsecutiveCalls(
                        returnValue(firstCommit),
                        returnValue(secondCommit)
                )
        );
        mockMultipleBazaarRepositoryManager
                .expects(exactly(2))
                .method("getRepository")
                .with(eq(1L))
                .will(returnValue(bzrMgr));

        Map entriesMap;
        List entries;
        revisionIndexer = getRevisionIndexer();

        entriesMap = revisionIndexer.getLogEntriesByRepository(issue);
        assertEquals(1, entriesMap.size());

        entries = (List) entriesMap.get(1L); /* Get entries from repository with ID 1 */

        assertEquals(2, entries.size());
        /* Bigger revision numbers on top */
        assertSame(firstCommit, entries.get(1));
        assertSame(secondCommit, entries.get(0));
    }

    public void testGetLogEntriesByProjectWhenUserHasNoPermissionToViewAnyIssuesMentionedInBzrCommit() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        IndexWriter indexWriter;
        Document doc;
        RevisionIndexer revisionIndexer;

        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        /* This is the first matching document with revision 1 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_PROJECTKEY, "TST", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_DATE, DateField.dateToString(new Date()), Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        indexWriter.close();

        mockIssue = new Mock(MutableIssue.class);
        issue = (Issue) mockIssue.proxy();

        mockIssueManager.expects(once()).method("getIssueObject").with(eq("TST-1")).will(returnValue(issue));
        mockPermissionManager.expects(once()).method("hasPermission")
                .with(eq(Permissions.VIEW_VERSION_CONTROL), same(issue), new IsAnything())
                .will(returnValue(false));

        Map entriesMap;

        revisionIndexer = getRevisionIndexer();
        entriesMap = revisionIndexer.getLogEntriesByProject("TST", null, Integer.MAX_VALUE);

        assertEquals(0, entriesMap.size());
    }

    public void testGetLogEntriesByProjectWhenUserHasPermissionToViewTheIssuesMentionedInBzrCommitButRevisionCannotBeFoundInAnyRepository() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        IndexWriter indexWriter;
        Document doc;
        RevisionIndexer revisionIndexer;

        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        /* This is the first matching document with revision 1 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_PROJECTKEY, "TST", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_DATE, DateField.dateToString(new Date()), Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        indexWriter.close();

        mockIssue = new Mock(MutableIssue.class);
        issue = (Issue) mockIssue.proxy();

        mockIssueManager.expects(once()).method("getIssueObject").with(eq("TST-1")).will(returnValue(issue));
        mockPermissionManager.expects(once()).method("hasPermission")
                .with(eq(Permissions.VIEW_VERSION_CONTROL), same(issue), new IsAnything())
                .will(returnValue(true));

        mockBzrMgr.expects(once()).method("getLogEntry").with(eq(BazaarRevision.getRevision(1))).will(returnValue(null));
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRepository").with(eq(1L)).will(returnValue(bzrMgr));

        Map entriesMap;

        revisionIndexer = getRevisionIndexer();
        entriesMap = revisionIndexer.getLogEntriesByProject("TST", null, Integer.MAX_VALUE);

        assertEquals(0, entriesMap.size());
    }

    public void testGetLogEntriesByProjectWhenEverythingIsOk() throws IOException, IndexException {
        Mock mockIssue;
        Issue issue;
        IndexWriter indexWriter;
        Document doc;
        RevisionIndexer revisionIndexer;
        IBazaarLogMessage bzrLogEntry;

        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        /* This is the first matching document with revision 1 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_PROJECTKEY, "TST", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_DATE, DateField.dateToString(new Date()), Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        indexWriter.close();

        mockIssue = new Mock(MutableIssue.class);
        issue = (Issue) mockIssue.proxy();

        mockIssueManager.expects(once()).method("getIssueObject").with(eq("TST-1")).will(returnValue(issue));
        mockPermissionManager.expects(once()).method("hasPermission")
                .with(eq(Permissions.VIEW_VERSION_CONTROL), same(issue), new IsAnything())
                .will(returnValue(true));

        bzrLogEntry = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "fixed TST-1", null, null, null, null);
        mockBzrMgr.expects(once()).method("getLogEntry").with(eq(BazaarRevision.getRevision(1))).will(returnValue(bzrLogEntry));
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRepository").with(eq(1L)).will(returnValue(bzrMgr));

        Map entriesMap;
        List entries;

        revisionIndexer = getRevisionIndexer();
        entriesMap = revisionIndexer.getLogEntriesByProject("TST", null, Integer.MAX_VALUE);
        entries = (List) entriesMap.get(1L);

        assertEquals(1, entriesMap.size()); /* There should be 1 entry because it matches repository 1 */
        assertEquals(1, entries.size()); /* There should be 1 entry because it matches revision 1 */
        assertSame(bzrLogEntry, entries.get(0));
    }

    public void testGetLogEntriesByVersionWhenUserHasNoPermissionToViewAnyIssuesOfVersionSpecified() throws IOException, IndexException {
        Mock mockIssue;
        Issue issueObject;
        GenericValue issue;
        IndexWriter indexWriter;
        Document doc;

        issue = new MockGenericValue("Issue", EasyMap.build("key", "TST-1"));

        mockVersionManager.expects(once()).method("getFixIssues").withAnyArguments().will(returnValue(EasyList.build( issue )));
        mockVersionManager.expects(once()).method("getAffectsIssues").withAnyArguments().will(returnValue(EasyList.build( issue )));



        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        /* This is the first matching document with revision 1 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_DATE, DateField.dateToString(new Date()), Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        indexWriter.close();

        mockIssue = new Mock(MutableIssue.class);
        issueObject = (Issue) mockIssue.proxy();

        mockIssueManager.expects(once()).method("getIssueObject").with(eq("TST-1")).will(returnValue(issueObject));
        mockPermissionManager.expects(once()).method("hasPermission")
                .with(eq(Permissions.VIEW_VERSION_CONTROL), same(issueObject), new IsAnything())
                .will(returnValue(false));
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRepository").with(eq(1L)).will(returnValue(bzrMgr));

        Map entriesMap;

        RevisionIndexer revisionIndexer = getRevisionIndexer();
        entriesMap = revisionIndexer.getLogEntriesByVersion((Version) new Mock(Version.class).proxy(), null, Integer.MAX_VALUE);

        assertEquals(0, entriesMap.size()); /* No matches */
    }

    public void testGetLogEntriesByVersionWhenEverythingIsOk() throws IOException, IndexException {
        Mock mockIssue;
        Issue issueObject;
        GenericValue issue;
        IndexWriter indexWriter;
        Document doc;
        IBazaarLogMessage bzrLogEntry;

        issue = new MockGenericValue("Issue", EasyMap.build("key", "TST-1"));

        mockVersionManager.expects(once()).method("getFixIssues").withAnyArguments().will(returnValue(EasyList.build( issue )));
        mockVersionManager.expects(once()).method("getAffectsIssues").withAnyArguments().will(returnValue(EasyList.build( issue )));



        /* Populate the search index */
        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);

        /* This is the first matching document with revision 1 */
        doc = new Document();
        doc.add(new Field(RevisionIndexer.FIELD_ISSUEKEY, "TST-1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REPOSITORY, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_REVISIONNUMBER, "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
        doc.add(new Field(RevisionIndexer.FIELD_DATE, DateField.dateToString(new Date()), Field.Store.YES, Field.Index.UN_TOKENIZED));
        indexWriter.addDocument(doc);

        indexWriter.close();

        mockIssue = new Mock(MutableIssue.class);
        issueObject = (Issue) mockIssue.proxy();

        mockIssueManager.expects(once()).method("getIssueObject").with(eq("TST-1")).will(returnValue(issueObject));
        mockPermissionManager.expects(once()).method("hasPermission")
                .with(eq(Permissions.VIEW_VERSION_CONTROL), same(issueObject), new IsAnything())
                .will(returnValue(true));

        bzrLogEntry = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "bar TST-1", null, null, null, null);
        mockBzrMgr.expects(once()).method("getLogEntry").with(eq(BazaarRevision.getRevision(1))).will(returnValue(bzrLogEntry));
        mockMultipleBazaarRepositoryManager.expects(once()).method("getRepository").with(eq(1L)).will(returnValue(bzrMgr));

        Map entriesMap;
        List entries;

        RevisionIndexer revisionIndexer = getRevisionIndexer();
        entriesMap = revisionIndexer.getLogEntriesByVersion((Version) new Mock(Version.class).proxy(), null, Integer.MAX_VALUE);
        entries = (List) entriesMap.get(1L);

        assertEquals(1, entriesMap.size()); /* There should be 1 entry because it matches repository 1 */
        assertEquals(1, entries.size()); /* There should be 1 entry because it matches revision 1 */
        assertSame(bzrLogEntry, entries.get(0));
    }

    public void testAddRepository() throws IOException, IndexException {
        /* We'll start off with am empty index */
        IndexWriter indexWriter;
        IndexReader indexReader;
        RevisionIndexer revisionIndexer;
        IBazaarLogMessage entry;

        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);
        indexWriter.close();

        /* Create a BazaarManager which will contain one entry to be indexed */
        entry = new CommandLineLogMessage("1", "dchui", "branch_nick", Util.format(new Date()), "foo-bar TST-1", null, null, null, null);
        mockBzrMgr.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBzrMgr.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
        mockBzrMgr.expects(once()).method("getLogEntries").withAnyArguments().will(returnValue(EasyList.build(entry)));
        mockBzrMgr.expects(atLeastOnce()).method("getDisplayName").withNoArguments();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(exactly(2)).method("getRepositoryList").withNoArguments().will(
                onConsecutiveCalls(
                        returnValue(Collections.EMPTY_LIST),
                        returnValue(EasyList.build(bzrMgr))
                )
        );

        revisionIndexer = new RevisionIndexer(multipleBazaarRepositoryManager,
                applicationProperties, versionManager,
		issueManager, permissionManager, changeHistoryManager, indexPathManager) {
            protected boolean isKeyInString(IBazaarLogMessage logEntry) { return true; }
            protected String getProjectKeyFromIssueKey(String issueKey) { return "TST"; }
            protected List<String> getIssueKeysFromString(IBazaarLogMessage logEntry) { return EasyList.build("TST-1"); }
        };
        revisionIndexer.addRepository(bzrMgr);

        indexReader = LuceneUtils.getIndexReader(getIndexPath());

        final TermDocs termDocs = indexReader.termDocs(new Term(RevisionIndexer.FIELD_REPOSITORY, Long.toString(1L)));

        assertTrue(termDocs.next()); /* Contains one entry only */
        assertFalse(termDocs.next()); /* Should return false if there is only one entry */

        indexReader.close();
    }

    public void testRemoveEntries() throws IOException, IndexException, BazaarClientException {
        /* We'll start off with am empty index */
        IndexWriter indexWriter;
        IndexReader indexReader;
        RevisionIndexer revisionIndexer;
        IBazaarLogMessage entry;

        indexWriter = LuceneUtils.getIndexWriter(getIndexPath(), true, RevisionIndexer.ANALYZER);
        indexWriter.close();

        /* Create a BazaarManager which will contain one entry to be indexed */
        entry = new CommandLineLogMessage("1", "dchui", "branch_nick", new Date().toString(), "foo TST-1", null, null, null, null);
        mockBzrMgr.expects(atMostOnce()).method("getRoot").withNoArguments().will(returnValue(System.getProperty("bzr.root")));
        mockBzrMgr.expects(atLeastOnce()).method("getId").withNoArguments().will(returnValue(1L));
        mockBzrMgr.expects(atLeastOnce()).method("isActive").withNoArguments().will(returnValue(true));
        mockBzrMgr.expects(atLeastOnce()).method("getLogEntries").withAnyArguments().will(returnValue(EasyList.build(entry)));
        mockBzrMgr.expects(atLeastOnce()).method("getDisplayName").withNoArguments();

        mockMultipleBazaarRepositoryManager.reset();
        mockMultipleBazaarRepositoryManager.expects(exactly(2)).method("getRepositoryList").withNoArguments().will(
                onConsecutiveCalls(
                        returnValue(EasyList.build(bzrMgr)), /* During the construction of the RevisionIndexer */
                        returnValue(EasyList.build(bzrMgr)) /* When RevisionIndexer.updateIndex() index is called */
                )
        );

        revisionIndexer = new RevisionIndexer(multipleBazaarRepositoryManager,
                applicationProperties, versionManager,
		issueManager, permissionManager, changeHistoryManager, indexPathManager) {
            protected boolean isKeyInString(IBazaarLogMessage logEntry) { return true; }
            protected String getProjectKeyFromIssueKey(String issueKey) { return "TST"; }
            protected List<String> getIssueKeysFromString(IBazaarLogMessage logEntry) { return EasyList.build("TST-1"); }
        };
        revisionIndexer.updateIndex();

        indexReader = LuceneUtils.getIndexReader(getIndexPath());
        assertEquals(1, indexReader.numDocs()); /* Contains one entry only. This is what we are going to change */
        indexReader.close();

        revisionIndexer.removeEntries(bzrMgr);
        indexReader = LuceneUtils.getIndexReader(getIndexPath());
        assertEquals(0, indexReader.numDocs()); /* There should be no matches */
        indexReader.close();
    }
}
