package it.com.atlassian.jira.plugin.ext.bazaar;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Properties;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.vcs.bazaar.client.BazaarClientFactory;
import org.vcs.bazaar.client.IBazaarClient;
import org.vcs.bazaar.client.core.BazaarClientException;

import com.atlassian.jira.webtests.JIRAWebTest;

/**
 * 
 * @author Guillermo Gonzalez <guillo.gonzo@gmail.com>
 *
 */
public class TestBazaarPlugin extends JIRAWebTest
{
    private String bzrRepositoryUrl;

    private IBazaarClient repository = null;

    private Properties testSubversionConfiguration;

    private void setupSubversionTestConfiguration() {
        InputStream in = null;

        try {
            testSubversionConfiguration = new Properties();

            in = getClass().getClassLoader().getResourceAsStream("subversion-jira-plugin.properties");
            assertNotNull(in);

            testSubversionConfiguration.load(in);

        } catch (final IOException ioe) {
            fail("Unable to get test BZR URL: " + ioe.getMessage());
        } finally {
            IOUtils.closeQuietly(in);
        }
    }

    public void setUp()
    {
        super.setUp();
        setupSubversionTestConfiguration();
        
        try
        {
            bzrRepositoryUrl = "file://" + testSubversionConfiguration.getProperty("bzr.test.root.path");
            assertNotNull(bzrRepositoryUrl);

            /* Just in case stuff is left over */
            removeBzrRepository();

            BazaarClientFactory.setupBestAvailableBackend(true);
            repository = BazaarClientFactory.createClient(BazaarClientFactory.getPreferredClientType());
        	repository.setWorkDir(new File(bzrRepositoryUrl));
        }
        catch (BazaarClientException e)
        {
            throw new RuntimeException("BZR error", e);
        }
        catch (IOException ioe)
        {
            throw new RuntimeException("Unable to remove BZR repository previously created by integeration tests at: " + bzrRepositoryUrl, ioe);
        }
        catch (URISyntaxException use)
        {
            throw new RuntimeException("Property bzr.root not setup to point to a local repository for tests.", use);
        }

        restoreData("TestBazaarPlugin.xml");
        activateSubversionRepository();
    }

    private void activateSubversionRepository()
    {
        gotoPage("/secure/ActivateSubversionRepository.jspa?repoId=1");
    }

    protected void removeBzrRepository() throws IOException, URISyntaxException {
        final File unzipDestination = new File(new URI(bzrRepositoryUrl));

        if (unzipDestination.exists())
            FileUtils.deleteDirectory(unzipDestination);
    }

    public void tearDown()
    {
        deleteSubversionRepository();

        try
        {
            createRemoveLogEntry("TearDown");
            removeBzrRepository();
        }
        catch (BazaarClientException e)
        {
            throw new RuntimeException("BZR error", e);
        }
        catch (IOException e)
        {
            throw new RuntimeException("Generic IO error", e);
        }
        catch (URISyntaxException use)
        {
            throw new RuntimeException("Invalid bzr.root specified in jira-subversion-plugin.properties.", use);
        }
        finally
        {
            super.tearDown();
        }
    }

    private void deleteSubversionRepository()
    {
        gotoPage("/secure/DeleteSubversionRepository!default.jspa?repoId=1");
        setWorkingForm("jiraform");
        submit("delete");
    }

    public TestBazaarPlugin(String string)
    {
        super(string);
    }

    /**
     * Anti test for <a href="http://jira.atlassian.com/browse/BZR-93">BZR-93</a>
     *
     * @throws BZRException
     * Thrown if the is a problem performing BZR operations.
     * @throws InterruptedException
     * Thrown if the wait for reindexing throws it.
     * @throws IOException 
     */
    public void testCommitWithJiraIssueKeysInLowerCase() throws BazaarClientException, InterruptedException, IOException
    {
        addProject("Another Homo", "AHSP", null, "admin", null);

        createIssueStep1("Another Homo", "Bug");
        setFormElement("summary", "A new issue");

        submit("Create"); /* Create another issue which will be mentioned in the commit message */


        createLogEntry("hsp-1, ahsp-1 A commit message with a valid JIRA issue key but in lower case form.");
        waitForReindex();

        /* Check at HSP-1 */
        gotoIssue("HSP-1");
        assertLinkPresentWithText("Subversion Commits");
        clickLinkWithText("Subversion Commits");

        assertTextPresent(
                "<a href=\"" + getEnvironmentData().getContext() + "/browse/HSP-1\" title=\"This is a test bug for the subversion plugin\">HSP-1</a>"
                        + ", <a href=\"" + getEnvironmentData().getContext() + "/browse/AHSP-1\" title=\"A new issue\">AHSP-1</a> "
                        + "A commit message with a valid JIRA issue key but in lower case form.");
        assertLinkPresentWithText("AHSP-1");

        /* Check at AHSP-1, and the subversion issue tab panel should already be selected. */
        gotoIssue("AHSP-1");

        assertTextPresent(
                "<a href=\"" + getEnvironmentData().getContext() + "/browse/HSP-1\" title=\"This is a test bug for the subversion plugin\">HSP-1</a>"
                        + ", <a href=\"" + getEnvironmentData().getContext() + "/browse/AHSP-1\" title=\"A new issue\">AHSP-1</a> "
                        + "A commit message with a valid JIRA issue key but in lower case form.");
        assertLinkPresentWithText("HSP-1");

        /* Check if the issues are rendered properly in the Subversion project tab panel */
        gotoPage("/browse/HSP?report=com.atlassian.jira.plugin.ext.bazaar:subversion-project-tab");
        assertTextPresent(
                "<a href=\"" + getEnvironmentData().getContext() + "/browse/HSP-1\" title=\"This is a test bug for the subversion plugin\">HSP-1</a>"
                        + ", <a href=\"" + getEnvironmentData().getContext() + "/browse/AHSP-1\" title=\"A new issue\">AHSP-1</a> "
                        + "A commit message with a valid JIRA issue key but in lower case form.");

        gotoPage("/browse/AHSP?report=com.atlassian.jira.plugin.ext.bazaar:subversion-project-tab");
        assertTextPresent(
                "<a href=\"" + getEnvironmentData().getContext() + "/browse/HSP-1\" title=\"This is a test bug for the subversion plugin\">HSP-1</a>"
                        + ", <a href=\"" + getEnvironmentData().getContext() + "/browse/AHSP-1\" title=\"A new issue\">AHSP-1</a> "
                        + "A commit message with a valid JIRA issue key but in lower case form.");
    }
    
    public void testIssueTabPanelForCommitMessages() throws BazaarClientException, InterruptedException, IOException
    {
        gotoIssue("HSP-1");

        // Browse and make sure nothing exists on the tab

        // Assert the Subversion tab exists
        assertLinkPresentWithText("Subversion Commits");

        clickLinkWithText("Subversion Commits");

        // Assert that no log entries exist
        //assertTextPresent("There are no subversion log entries for this issue yet.");

        // Connect to BZR and create a log entrywait
        createLogEntry("HSP-1 adding a bzr log message");

        createModifyLogEntry("HSP-1 a modify bzr log message");

        // NOTE: this is trying to create a log entry that will be marked as replace but the api is a bit unclear
        // as to how to do this, it does not work at the moment.
        //createReplaceLogEntry("HSP-1 a modify bzr log message");

        createRemoveLogEntry("HSP-1 a remove bzr log message");

        // Trigger a re-index in the plugin
        // wait for the system to re-index
        waitForReindex();

        // Borwse and verify the commit log exists
        gotoIssue("HSP-1");

        // Assert the Subversion tab exists
        assertTextPresent("Subversion Commits");

        // Assert that an add log entry exists
        assertTextPresent("adding a bzr log message");
        assertTextPresent("ADD");
        // Make sure that the two add viewcvs links are present and in the correct format
        assertTextPresent("http://bzr.atlassian.com/fisheye/changelog/public?cs=1"); /* Changeset  */
        assertTextPresent("http://bzr.atlassian.com/fisheye/viewrep/public/test/file.txt?r=1"); /* First file */
        assertTextPresent("http://bzr.atlassian.com/fisheye/viewrep/public/test?r=1"); /* Second file */

        // Assert that a modify log entry exists
        assertTextPresent("modify bzr log message");
        assertTextPresent("MODIFY");
        // Make sure that the modify viewcvs link exists
        assertTextPresent("http://bzr.atlassian.com/fisheye/viewrep/public/test/file.txt#2");

        // Assert that a delete log entry exists
        assertTextPresent("remove bzr log message");
        assertTextPresent("DEL");
        // Make sure that the del viewcvs link exists
        assertTextPresent("http://bzr.atlassian.com/fisheye/viewrep/public/test");
    }

    public void testIssueTabPanelWithSimilarProjectKeysInCommitMessages() throws BazaarClientException, InterruptedException, IOException
    {
        createCommitsWithSimilarProjectKeys();

        gotoPage("/browse/HSP-1?page=com.atlassian.jira.plugin.ext.bazaar:subversion-commits-tabpanel");

        assertTextPresent("a modify bzr log message");
        assertTextNotPresent("AHSP");
        assertTextNotPresent("This is an entry with a URL");

        gotoPage("/browse/AHSP-1?page=com.atlassian.jira.plugin.ext.bazaar:subversion-commits-tabpanel");
        assertTextNotPresent("a modify bzr log message");
        assertTextNotPresent("This is an entry with a URL");
        assertTextPresent("This is a commit for");

        createRemoveLogEntry("Removed.");
    }

    public void testProjectTabPanelWithSimilarProjectKeysInCommitMessages() throws BazaarClientException, InterruptedException, IOException {
        createCommitsWithSimilarProjectKeys();

        gotoPage("/browse/HSP?report=com.atlassian.jira.plugin.ext.bazaar:subversion-project-tab");
        assertTextPresent("a modify bzr log message");
        assertTextNotPresent("AHSP");
        assertTextNotPresent("This is an entry with a URL");

        gotoPage("/browse/AHSP?page=com.atlassian.jira.plugin.ext.bazaar:subversion-commits-tabpanel");
        assertTextNotPresent("a modify bzr log message");
        assertTextNotPresent("This is an entry with a URL");
        assertTextPresent("This is a commit for");

        createRemoveLogEntry("Removed.");
    }

    public void testFilterCommitMessagesInProjectTabPanelByVersion() throws BazaarClientException, InterruptedException, IOException {
    	final String versionId2;
    	final String versionId1;

		addProject("Test Project", "TST", null, "admin", null);
		versionId1 = addVersion("Test Project", "1.0", null);
		versionId2 = addVersion("Test Project", "2.0", null);

		addIssue("Test Project", "TST", ISSUE_TYPE_BUG, "Summary Of Bug 1", null, null,
				new String[] { "1.0" }, null, null, null, null, null, null, null);
		addIssue("Test Project", "TST", ISSUE_TYPE_BUG, "Summary Of Bug 2", null, null,
				new String[] { "2.0" }, null, null, null, null, null, null, null);

        createLogEntry("This is a commit for TST-1.");
        createModifyLogEntry("This is a commit for TST-2.");

        waitForReindex();

        gotoPage("/browse/TST?report=com.atlassian.jira.plugin.ext.bazaar:subversion-project-tab");
        assertTextPresent("This is a commit for <a href=\"" + getEnvironmentData().getContext() + "/browse/TST-1\" title=\"Summary Of Bug 1\">TST-1</a>");
        assertTextPresent("This is a commit for <a href=\"" + getEnvironmentData().getContext() + "/browse/TST-2\" title=\"Summary Of Bug 2\">TST-2</a>");

        gotoPage("/browse/TST?report=com.atlassian.jira.plugin.ext.bazaar:subversion-project-tab&selectedVersion=" + versionId1);
        assertTextPresent("This is a commit for <a href=\"" + getEnvironmentData().getContext() + "/browse/TST-1\" title=\"Summary Of Bug 1\">TST-1</a>");
        assertTextNotPresent("This is a commit for <a href=\"" + getEnvironmentData().getContext() + "/browse/TST-2\" title=\"Summary Of Bug 2\">TST-2</a>");

        gotoPage("/browse/TST?report=com.atlassian.jira.plugin.ext.bazaar:subversion-project-tab&selectedVersion=" + versionId2);
        assertTextNotPresent("This is a commit for <a href=\"" + getEnvironmentData().getContext() + "/browse/TST-1\" title=\"Summary Of Bug 1\">TST-1</a>");
        assertTextPresent("This is a commit for <a href=\"" + getEnvironmentData().getContext() + "/browse/TST-2\" title=\"Summary Of Bug 2\">TST-2</a>");
    }

	protected void createCommitsWithSimilarProjectKeys() throws BazaarClientException,
			InterruptedException, IOException {
        addProject("Another Homo", "AHSP", null, "admin", null);

        createIssueStep1("Another Homo", "Bug");
        setFormElement("summary", "A new issue");

        submit("Create");


        createLogEntry("This is a commit for AHSP-1.");
        createModifyLogEntry("HSP-1 a modify bzr log message");
        createModifyLogEntry("This is an entry with a URL http://jira.atlassian.com/browse/HSP-1");

        waitForReindex();
	}

    private void createRemoveLogEntry(String message) throws BazaarClientException
    {
    	deleteDir("test");
    }

    private void createModifyLogEntry(String message) throws BazaarClientException
    {
        byte[] modifiedContents = "This is the same file but modified a little.".getBytes();

        modifyFile("test", "test/file.txt", modifiedContents);
    }


    private void waitForReindex() throws InterruptedException
    {
        Thread.sleep(150000);
    }

    private void createLogEntry(String message) throws BazaarClientException, IOException
    {

        byte[] contents = "This is a new file".getBytes();
        addDir("test", "test/file.txt", contents);
    }

    /*
    * This method performs commiting an addition of a  directory  containing  a
    * file.
    */
    private void addDir(String dirPath,
                                 String filePath, byte[] data) throws BazaarClientException, IOException
    {
       File dir = new File(dirPath);
       dir.createNewFile();
       File file = new File(filePath);
       file.createNewFile();
       repository.add(new File[]{new File(dir, file.getName())});
    }

    /*
    * This method performs committing file modifications.
    */
    private void modifyFile(String dirPath,
                                     String filePath, byte[] newData) throws BazaarClientException
    {
    	
    }

    /*
    * This method performs committing a deletion of a directory.
    */
    private void deleteDir(String dirPath) throws BazaarClientException
    {
    	repository.remove(new File[]{new File(dirPath)});
    }

}
